from random import random
from hgtk.checker import has_batchim
from ..utils.dictionary import sort_by_key
from ..utils.hangul import create_possible_syllables, is_complete_syllable

# See https://namu.wiki/w/야민정음
raw_yamin_dictionary = {
    # 한글 - 한글 (받침 포함)
    '겨': '저',
    '대': '머',
    '귀': '커',
    '파': '과',
    '피': '끠',
    '끠': '괴',
    '괴': '피',
    '포': '쪼',
    '비': '네',
    '삐': '볘',
    '눠': '부',
    '며': '띠',
    '거': '지',
    '고': '끄',
    '켜': '궈',
    '교': '꼬',
    '티': '더',

    # 한글 - 한글 (받침 미포함)
    '인': '외',
    '유': '윾',
    '위': '읶',
    '우': '윽',
    '굿': '긋',
    '공': '퐁',
    '읍': '윰',
    '왕': '앟',
    '왱': '앻',
    '왓': '앛',
    '왯': '앷',
    '욋': '잋',
    '통': '듷',
    '봉': '넣',
    '두': '득',
    '딘': '되',
    '익': '의',
    '임': '읜',
    '일': '읟',
    '억': '여',
    '백': '뿌',
    '산': '솨',
    '번': '놴',
    '부': '넉',
    '보': '넌',

    '근': 'ㄹ',
    '너': 'ㅂ',
    '든': 'ㅌ',
    '병': 'ㅹ',

    '개새끼': '7H^H77l',
    '대통령': '새토깽',
    '이명박': '어맹뿌',
    '푸틴': '곡틴',

    # 한글 - 영어
    '내': 'LH',
    '배': 'IdH',
    '애': 'OH',
    '야': 'OF',
    '태': 'EH',

    # 한글 - 한자
    '뉘': '爿',
    '몸': '啚',
    '묵': '号',
    '숲': '金',
    '스': '亼',
    '슥': '今',
    '쓰': '丛',
    '조': '丕',
    '크': '彐',
    '튽': '長',
    '표': '丑',
    '푸': '辛',

    # 한글 - 기타문자
    '망': 'ㅁ5',
    '응': '%',
    'ㅋ': 'ヲ',

    # 영어 - 영어
    'I': 'l',
    'cl': 'd',
    'lo': 'b',
    'rn': 'm',

    # 영어 - 기타문자
    'l': '|',
    '|': 'I',
    'EU': '日J',

    # 기타
    '4': 'Ч',

    # 글자 압축
    '구구': '뀨',
    '굴국': '뀱',
    '도도': '뚀',
    '돌돔': '뚊',
    '딘딘': '田',
    '보보': '뾰',
    '복복': '뾲',
    '부부': '쀼',
    '북북': '쀾',
    '불붙': '쁉',
    'ㅅㅂㄹㅁ': '섊',
    '속삭': '쏶',
    '스스': '쓰',
    '스시': '씌',
    '조조': '쪼',
    '존좋': '쬲',
    '주작': '짞',
    '주주': '쮸',

    # 180도 회전
    '근육': '뇽근',
    '눈물': '롬곡',
    '물음표': '표믕롬',
    '육군': '곤뇽',
    '폭풍': '옾눞',
}

yamin_dictionary = raw_yamin_dictionary.copy()

for key, value in raw_yamin_dictionary.items():
    if is_complete_syllable(key) and not has_batchim(key) \
            and is_complete_syllable(value) and not has_batchim(value):
        first_chars = create_possible_syllables(key)
        second_chars = create_possible_syllables(value)
        for i in range(27):
            yamin_dictionary[first_chars[i]] = second_chars[i]

for key, value in yamin_dictionary.copy().items():
    if value not in yamin_dictionary:
        yamin_dictionary[value] = key

yamin_dictionary = sort_by_key(yamin_dictionary, lambda a, b:
                               (1 if len(a[0]) < len(b[0]) else -1) if len(a[0]) != len(b[0])
                               else (-1 if a[0] < b[0] else 1))


def encode_yamin(text: str, *, active_rate=1.0) -> str:
    yamin_pairs = yamin_dictionary.items()
    for idx, (key, _) in enumerate(yamin_pairs):
        if random() < active_rate:
            text = text.replace(key, f'$__{idx}__$')
    for idx, (_, value) in enumerate(yamin_pairs):
        text = text.replace(f'$__{idx}__$', value)
    return text



from .rules import PhonemeReplaceRule
from ..utils.hangul import split_by_phoneme, join_phonemes, join_syllables


def replace_phoneme(text: str, *, rule: PhonemeReplaceRule) -> str:
    phonemes = split_by_phoneme(text)
    for index, recipe in enumerate(phonemes):
        phonemes[index] = rule.apply(recipe, phonemes, index)

    chars = join_phonemes(phonemes)
    text = join_syllables(chars)
    return text



from typing import List
from ..utils.hangul import CharRecipe, has_phoneme, replace_phoneme


class PhonemeReplaceRule:
    def apply(self, recipe: CharRecipe, recipes: List[CharRecipe], index: int) -> CharRecipe:
        pass


class SquareDog(PhonemeReplaceRule):
    """
    담신믄 네모네모 멈뭄미뫄 눈미 마주치고 말맜습니다.
    담신믄 미제 네모네모 멈뭄미믜 저주로 돔그란 글자를 칠 수 멊습니다. 멈멈!
    """

    def apply(self, recipe: CharRecipe, recipes: List[CharRecipe], index: int) -> CharRecipe:
        return replace_phoneme(recipe, 'ㅇ', 'ㅁ') if has_phoneme(recipe, 'ㅇ') else recipe


class CircleDog(PhonemeReplaceRule):
    """
    짠! 당신의 저주는 풀렸습니다!
    하지안 이제부터 당신은 동글동글 엉엉이와 눈이 아주치고 알았습니다!
    당신은 이제 동글동글 엉엉이의 저주로 네오난 글자를 칠 수 없습니다. 엉엉!
    """

    def apply(self, recipe: CharRecipe, recipes: List[CharRecipe], index: int) -> CharRecipe:
        return replace_phoneme(recipe, 'ㅁ', 'ㅇ') if has_phoneme(recipe, 'ㅁ') else recipe


class UShapedDog(PhonemeReplaceRule):
    """
    빠밤! 당신의 저주는 풀렸습니다!
    하지반 답신븐 법붑비 2협제봐 눈비 바주쳤습니다!
    답신븐 비제 법붑비 2협제븨 저주로 네보난 글자봐 돕그란 글자를 칠 수 벖습니다. 법법!
    """

    def apply(self, recipe: CharRecipe, recipes: List[CharRecipe], index: int) -> CharRecipe:
        return replace_phoneme(replace_phoneme(recipe, 'ㅁ', 'ㅂ'), 'ㅇ', 'ㅂ') \
            if has_phoneme(recipe, 'ㅁ') or has_phoneme(recipe, 'ㅇ') else recipe


class SeaoCat(PhonemeReplaceRule):
    """
    안녕하새오 고양이애오 겨울 추어오 문 열어주새오
    가족 대려오개 감사해오 문 열어주새오
    """

    def apply(self, recipe: CharRecipe, recipes: List[CharRecipe], index: int) -> CharRecipe:
        if has_phoneme(recipe, 'ㅔ'):
            return replace_phoneme(recipe, 'ㅔ', 'ㅐ')
        elif has_phoneme(recipe, 'ㅖ'):
            return replace_phoneme(recipe, 'ㅖ', 'ㅐ')
        elif has_phoneme(recipe, 'ㅒ'):
            return replace_phoneme(recipe, 'ㅒ', 'ㅐ')
        elif has_phoneme(recipe, 'ㅝ'):
            return replace_phoneme(recipe, 'ㅝ', 'ㅓ')
        elif has_phoneme(recipe, 'ㅞ'):
            return replace_phoneme(recipe, 'ㅞ', 'ㅓ')
        elif has_phoneme(recipe, 'ㅛ'):
            return replace_phoneme(recipe, 'ㅛ', 'ㅗ')
        return recipe



from random import random
from hgtk.letter import decompose
from ..utils.hangul import is_complete_syllable


def shred_syllable(text: str, *, active_rate=0.1) -> str:
    output = ''
    for char in text:
        if random() < active_rate and is_complete_syllable(char):
            output += ''.join(decompose(char))
        else:
            output += char
    return output



from typing import List, Tuple, Union
from hgtk.const import JONG
from hgtk.letter import compose, decompose

CharRecipe = Union[Tuple[str, str, str], str]


def has_phoneme(recipe: CharRecipe, target_phoneme: str) -> bool:
    return target_phoneme in recipe if isinstance(recipe, tuple) else recipe == target_phoneme


def replace_phoneme(recipe: CharRecipe, old_phoneme: str, new_phoneme: str) -> str:
    if isinstance(recipe, tuple):
        return tuple([new_phoneme if phoneme == old_phoneme else phoneme for phoneme in recipe])
    return new_phoneme if recipe == old_phoneme else old_phoneme


def split_by_phoneme(text: str) -> List[CharRecipe]:
    return [decompose(char) if is_complete_syllable(char) else char for char in text]


def join_phonemes(phonemes: List[CharRecipe]) -> List[str]:
    return [compose(*recipe) if isinstance(recipe, tuple) else recipe for recipe in phonemes]


def create_possible_syllables(char: str) -> List[str]:
    recipe = decompose(char)
    characters = [compose(recipe[0], recipe[1], jong) for jong in JONG[1:]]
    return characters


def is_complete_syllable(char: str) -> bool:
    FRIST_HANGUL_UNICODE = 0xAC00
    LAST_HANGUL_UNICODE = 0xD7A3
    return len(char) == 1 and FRIST_HANGUL_UNICODE <= ord(char) <= LAST_HANGUL_UNICODE


def join_syllables(chars: List[str]) -> str:
    return ''.join(chars)



from codecs import open
from os import path
from setuptools import setup, find_packages

basedir = path.abspath(path.dirname(__file__))
with open(path.join(basedir, 'README.md'), encoding='utf-8') as file:
    long_description = file.read().replace('\r\n', '\n')

setup(
    name='kotka',
    version='0.1.1',
    description='Korean Obfuscation ToolKit Advanced',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Seungjae Park',
    author_email='astro.psj@gmail.com',
    url='https://github.com/Astro36/kotka',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Korean',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Linguistic',
    ],
    license='MIT',
    keywords='korean-nlp, korean-text-processing, hangul, toolkit',
    python_requires='>=3.6',
)
